#include "aneu_meshloader.h"
#include <fstream>
#include <algorithm>
#include <set>
#include <stdexcept>
#include <iterator>

Node AneuMeshLoader::interpolatedNode(Node& node_1, Node& node_2)
{
    int dimensions = node_1.m_coords.size();
    if (dimensions != node_2.m_coords.size())
        throw std::runtime_error("Nodes' dimensions don't match");
    Node result_node;
    for (int i = 0; i != dimensions; i++)
        result_node.m_coords.push_back((node_1.m_coords[i] + node_2.m_coords[i]) / 2.0f);
    return result_node;
}

void AneuMeshLoader::loadMesh(char* p) //!!! Мы давно std::string используем
{
    std::ifstream geom_file(p);
    if (!geom_file)
        throw std::runtime_error("Error loading geomentry file");
    int quant, param, counter;
    geom_file >> quant >> param;
    m_nodes.reserve(quant);
    for (int i = 0; i < quant; i++)
    {
        Node new_node;
        new_node.m_ID = i;
        new_node.m_coords.reserve(param);
        float coord;
        for (int j = 0; j != param; j++)
        {
            geom_file >> coord;
            new_node.m_coords.push_back(coord);
        }
        m_nodes.push_back(new_node);
    }

    geom_file >> quant >> param;
    m_elements.reserve(quant);
    for (int i = 0; i != quant; i++)
    {
        Element new_element;
        int node_ID;
        new_element.m_ID = i;
        geom_file >> new_element.m_mat_ID;
        for (int j = 0; j != param; j++)
        {
            geom_file >> node_ID;
            new_element.m_nodes.push_back(node_ID - 1);
        }
        m_elements.push_back(new_element);
    }

    geom_file >> quant >> param;
    m_surfaces.reserve(quant);
    for (int i = 0; i != quant; i++)
    {
        Surface new_surface;
        int node_ID;
        geom_file >> new_surface.m_ID;
        for (int j = 0; j != param; j++)
        {
            geom_file >> node_ID;
            new_surface.m_nodes.push_back(node_ID - 1);
        }
        m_surfaces.push_back(new_surface);
    }

    geom_file.close();
}

std::vector<Node>& AneuMeshLoader::getNodes()
{
    return m_nodes;
}

std::vector<Element>& AneuMeshLoader::getElements()
{
    return m_elements;
}

std::vector<Surface>& AneuMeshLoader::getSurfaces()
{
    return m_surfaces;
}

std::vector<Element> AneuMeshLoader::findElem(int id_1, int id_2)
{
    std::vector<Element> found_elements;
    auto   last = m_elements.begin();
    while (last != m_elements.end())
    {
        last = find_if(last, m_elements.end(), [](const Element& elem) //!!!                             Вы elem изменяете внутри лямбды? Почему без const 
            {
                return (std::find(elem.m_nodes.begin(), elem.m_nodes.end(), id_1) != elem.m_nodes.end()
                    && std::find(elem.m_nodes.begin(), elem.m_nodes.end(), id_2) != elem.m_nodes.end());
            });

        if (last != m_elements.end())
            found_elements.push_back(*last++);
    }
    return found_elements;
}

std::vector<Element> AneuMeshLoader::findElem(int id_1, int id_2, int id_3)
{
    std::vector<Element> found_elements;
    std::vector<Element>::iterator last = m_elements.begin();
    while (last != m_elements.end())
    {
        last = find_if(last, m_elements.end(), [](const Element& elem) //!!!                              Вы elem изменяете внутри лямбды? Почему без const
            {
                return (std::find(elem.m_nodes.begin(), elem.m_nodes.end(), id_1) != elem.m_nodes.end()
                    && std::find(elem.m_nodes.begin(), elem.m_nodes.end(), id_2) != elem.m_nodes.end()
                    && std::find(elem.m_nodes.begin(), elem.m_nodes.end(), id_3) != elem.m_nodes.end());
            });
        if (last != m_elements.end())
            found_elements.push_back(*last++);
    }
    return found_elements;
}

void AneuMeshLoader::getNodesBySurfaceID(int sur_ID, std::vector<int>& container)
{
    //!!! ЛР расчитана наиспользрвание алгорритмов, а не циклов for. Это касается всех методов.
    
    for (const auto& s : m_surfaces)
        if (s.m_ID == sur_ID)
            for (auto& n : s.m_nodes)
                container.push_back(n);
                
    //!!! Это слишком долго. Нужно выбрать правильный тип контейнера, чтобы не приходилось дубликаты убирать
    
    std::sort(container.begin(), container.end());
    container.erase(std::unique(container.begin(), container.end()), container.end());
}

void AneuMeshLoader::getElementsByMatID(int mat_ID, std::vector<Element>& container)
{
    for (auto& e : m_elements)  //!!! Через алгоритм
        if (e.m_mat_ID == mat_ID)
            container.push_back(e);
}

void AneuMeshLoader::getSurfacesByID(int sur_ID, std::vector<Surface>& container)
{
    for (auto& s : m_surfaces) //!!! Через алгоритм
        if (s.m_ID == sur_ID)
            container.push_back(s); //!!! Вектор будет довольно часто заново выделять память
}

void AneuMeshLoader::quadrateMeshElements()
{
    int nodes = m_elements.begin()->m_nodes.size();
    
    for (auto& e : m_elements)
    {
        //!!! У нас на одном ребре множество тетраэдров.
        //!!! И как вы учитываете, что обходя данный тетраэдр у него могут быть ребра, на которые мы уже добавили узел?
        
        for (int i = 0; i != nodes; i++)
            for (int j = i + 1; j != nodes; j++)
            {
                Node new_node = interpolatedNode(m_nodes[e.m_nodes[i]], m_nodes[e.m_nodes[i + 1]]);
                e.m_nodes.push_back(new_node.m_ID = m_nodes.size());
                m_nodes.push_back(new_node);
            }
    }
}

void AneuMeshLoader::getNeighbors(std::vector<std::vector<int>>& container)
{
    //!!! Полагается, что мы работает с большими данным: узлов и КЭ больше 100 млн.
    //!!! Читайте задание повнимательнее - обеспечить сложность не более логарифмической.
    //!!! У вас тут O(N * E) - для каждого узла обходятся все элементы...
    //!!! Тип внутреннего контейнера выбран неверно.
    
    for (auto& n : m_nodes) 
    {
        std::vector<int> node_neighbors;
        for (auto& e : m_elements)
            if (std::find(e.m_nodes.begin(), e.m_nodes.end(), n.m_ID + 1) != e.m_nodes.end())
                node_neighbors.insert(node_neighbors.end(), e.m_nodes.begin(), e.m_nodes.end());

        std::sort(node_neighbors.begin(), node_neighbors.end());
        node_neighbors.erase(std::unique(node_neighbors.begin(), node_neighbors.end()), node_neighbors.end());
        container.push_back(node_neighbors);
    }
}

std::ostream& operator<<(std::ostream& stream, AneuMeshLoader& mesh)
{
    stream << "Nodes:" << std::endl;
    for (auto& n : mesh.getNodes())
    {
        stream << "ID: " << n.m_ID << "Coords: ";
        for (auto& c : n.m_coords) stream << ' ' << c;
        stream << std::endl;
    }

    stream << "Elements:" << std::endl;
    for (auto& e : mesh.getElements())
    {
        stream << "ID: " << e.m_ID << "Material ID: " << e.m_mat_ID << "Nodes' IDs: ";
        for (auto& n : e.m_nodes) stream << ' ' << n;
        stream << std::endl;
    }

    stream << "Surfaces:" << std::endl;
    for (auto& s : mesh.getSurfaces())
    {
        stream << "ID: " << s.m_ID << "Nodes' IDs: ";
        for (auto& n : s.m_nodes) stream << ' ' << n;
        stream << std::endl;
    }

    return stream;
}
